# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import division

import numpy as np
import pytest

import odl
from odl.trafos.fourier import (
    DiscreteFourierTransform, DiscreteFourierTransformInverse,
    FourierTransform)
from odl.trafos.util.ft_utils import (
    _interp_kernel_ft, dft_postprocess_data, dft_preprocess_data,
    reciprocal_grid)
from odl.core.util import (
    all_almost_equal, complex_dtype, conj_exponent, is_real_dtype,
    noise_element, skip_if_no_pyfftw)
from odl.core.util.testutils import simple_fixture

from odl.core.array_API_support import allclose

# --- pytest fixtures --- #


impl = simple_fixture(
    'impl',
    [pytest.param('numpy'),
     pytest.param('pyfftw', marks=skip_if_no_pyfftw),
     pytest.param('default')]
)
exponent = simple_fixture('exponent', [2.0, 1.0, float('inf'), 1.5])
sign = simple_fixture('sign', ['-', '+'])


# --- Helper functions --- #


def _params_from_dtype(dtype):
    if is_real_dtype(dtype):
        halfcomplex = True
    else:
        halfcomplex = False
    return halfcomplex, complex_dtype(dtype)


@pytest.fixture
def _dft_complex_space(odl_complex_floating_dtype, odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    shape = (4,5)
    return odl.uniform_discr(
        [0] * 2,
        np.subtract(shape, 1),
        shape,
        dtype=odl_complex_floating_dtype,
        nodes_on_bdry=True,
        impl=impl, device=device
    )
def skip_incompatible_impl(impl, dft_space):
    if dft_space.impl != 'numpy' and impl!= 'default':
        pytest.skip('Currently, only the numpy backend supports fft implementations that are not array-API compatible (e.g pyfftw)')

def _dft_space(shape, dtype='float64'):
    try:
        ndim = len(shape)
    except TypeError:
        ndim = 1
    return odl.uniform_discr(
        [0] * ndim,
        np.subtract(shape, 1),
        shape,
        dtype=dtype,
        nodes_on_bdry=True,
    )


def sinc(x):
    # numpy.sinc scales by pi, we don't want that
    return np.sinc(x / np.pi)

def sinc_pytorch(x):
    # torch.sinc scales by pi, we don't want that
    import torch 
    return torch.sinc(x / np.pi)


# ---- DiscreteFourierTransform ---- #


def test_dft_init(impl, odl_impl_device_pairs):
    # Just check if the code runs at all
    backend_impl, device = odl_impl_device_pairs
    shape = (4, 5)
    dom = _dft_space(shape)
    dom_nonseq = odl.uniform_discr([0, 0], [1, 1], shape, impl=backend_impl, device=device)
    dom_f32 = dom.astype('float32')
    ran = _dft_space(shape, dtype='complex128')
    ran_c64 = ran.astype('complex64')
    ran_hc = _dft_space((3, 5), dtype='complex128')

    # Implicit range
    DiscreteFourierTransform(dom, impl=impl)
    DiscreteFourierTransform(dom_nonseq, impl=impl)
    DiscreteFourierTransform(dom_f32, impl=impl)
    DiscreteFourierTransform(dom, axes=(0,), impl=impl)
    DiscreteFourierTransform(dom, axes=(0, -1), impl=impl)
    DiscreteFourierTransform(dom, axes=(0,), halfcomplex=True, impl=impl)
    DiscreteFourierTransform(dom, impl=impl, sign='+')

    # Explicit range
    DiscreteFourierTransform(dom, range=ran, impl=impl)
    DiscreteFourierTransform(dom_f32, range=ran_c64, impl=impl)
    DiscreteFourierTransform(dom, range=ran, axes=(0,), impl=impl)
    DiscreteFourierTransform(dom, range=ran, axes=(0,), impl=impl, sign='+')
    DiscreteFourierTransform(dom, range=ran, axes=(0, -1), impl=impl)
    DiscreteFourierTransform(dom, range=ran_hc, axes=(0,), impl=impl,
                             halfcomplex=True)


def test_dft_init_raise():
    # Test different error scenarios
    shape = (4, 5)
    dom = _dft_space(shape)
    dom_f32 = _dft_space(shape, dtype='float32')

    # Bad types
    with pytest.raises(TypeError):
        DiscreteFourierTransform(dom.tspace)

    with pytest.raises(TypeError):
        DiscreteFourierTransform(dom, dom.tspace)

    # Illegal arguments
    with pytest.raises(ValueError):
        DiscreteFourierTransform(dom, impl='fftw')

    with pytest.raises(ValueError):
        DiscreteFourierTransform(dom, axes=(1, 2))

    with pytest.raises(ValueError):
        DiscreteFourierTransform(dom, axes=(1, -3))

    # Badly shaped range
    bad_ran = _dft_space((3, 5), dtype='complex128')
    with pytest.raises(ValueError):
        DiscreteFourierTransform(dom, bad_ran)

    bad_ran = _dft_space((10, 10), dtype='complex128')
    with pytest.raises(ValueError):
        DiscreteFourierTransform(dom, bad_ran)

    bad_ran = _dft_space((4, 5), dtype='complex128')
    with pytest.raises(ValueError):
        DiscreteFourierTransform(dom, bad_ran, halfcomplex=True)

    bad_ran = _dft_space((4, 3), dtype='complex128')
    with pytest.raises(ValueError):
        DiscreteFourierTransform(dom, bad_ran, halfcomplex=True, axes=(0,))

    # Bad data types
    bad_ran = _dft_space(shape, dtype='complex64')
    with pytest.raises(ValueError):
        DiscreteFourierTransform(dom, bad_ran)

    bad_ran = _dft_space(shape, dtype='float64')
    with pytest.raises(ValueError):
        DiscreteFourierTransform(dom, bad_ran)

    bad_ran = _dft_space((4, 3), dtype='float64')
    with pytest.raises(ValueError):
        DiscreteFourierTransform(dom, bad_ran, halfcomplex=True)

    bad_ran = _dft_space((4, 3), dtype='complex128')
    with pytest.raises(ValueError):
        DiscreteFourierTransform(dom_f32, bad_ran, halfcomplex=True)

    # Bad sign
    with pytest.raises(ValueError):
        DiscreteFourierTransform(dom, sign=-1)


def test_dft_range():
    # 1d
    shape = 10
    dom = _dft_space(shape, dtype='complex128')
    fft = DiscreteFourierTransform(dom)
    true_ran = _dft_space(shape, dtype='complex128')
    assert fft.range == true_ran

    # 3d
    shape = (3, 4, 5)
    ran = _dft_space(shape, dtype='complex64')
    fft = DiscreteFourierTransform(ran)
    true_ran = _dft_space(shape, dtype='complex64')
    assert fft.range == true_ran

    # 3d, with axes and halfcomplex
    shape = (3, 4, 5)
    axes = (-1, -2)
    ran_shape = (3, 3, 5)
    dom = _dft_space(shape, dtype='float32')
    fft = DiscreteFourierTransform(dom, axes=axes, halfcomplex=True)
    true_ran = _dft_space(ran_shape, dtype='complex64')
    assert fft.range == true_ran


# ---- DiscreteFourierTransformInverse ---- #


def test_idft_init(impl):
    # Just check if the code runs at all; this uses the init function of
    # DiscreteFourierTransform, so we don't need exhaustive tests here
    shape = (4, 5)
    ran = _dft_space(shape, dtype='complex128')
    ran_hc = _dft_space(shape, dtype='float64')
    dom = _dft_space(shape, dtype='complex128')
    dom_hc = _dft_space((3, 5), dtype='complex128')

    # Implicit range
    DiscreteFourierTransformInverse(dom, impl=impl)

    # Explicit range
    DiscreteFourierTransformInverse(ran, domain=dom, impl=impl)
    DiscreteFourierTransformInverse(ran_hc, domain=dom_hc, axes=(0,),
                                    impl=impl, halfcomplex=True)


def test_dft_call(impl, _dft_complex_space):
    # 2d, complex, all ones and random back & forth
    # dft_dom = _dft_space(shape, dtype='complex64')
    skip_incompatible_impl(impl,_dft_complex_space)
    dft_dom = _dft_complex_space
    dft = DiscreteFourierTransform(domain=dft_dom, impl=impl)
    idft = DiscreteFourierTransformInverse(range=dft_dom, impl=impl)

    assert dft.domain == idft.range
    assert dft.range == idft.domain

    one = dft.domain.one()
    one_dft1 = dft(one, flags=('FFTW_ESTIMATE',))
    one_dft2 = dft.inverse.inverse(one, flags=('FFTW_ESTIMATE',))
    one_dft3 = dft.adjoint.adjoint(one, flags=('FFTW_ESTIMATE',))
    true_dft = [[20, 0, 0, 0, 0],  # along all axes by default
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0]]
    assert all_almost_equal(one_dft1, true_dft)
    assert all_almost_equal(one_dft2, true_dft)
    assert all_almost_equal(one_dft3, true_dft)

    one_idft1 = idft(one_dft1, flags=('FFTW_ESTIMATE',))
    one_idft2 = dft.inverse(one_dft1, flags=('FFTW_ESTIMATE',))
    one_idft3 = dft.adjoint(one_dft1, flags=('FFTW_ESTIMATE',))
    assert all_almost_equal(one_idft1, one)
    assert all_almost_equal(one_idft2, one)
    assert all_almost_equal(one_idft3, one)

    rand_arr = noise_element(dft_dom)
    rand_arr_dft = dft(rand_arr, flags=('FFTW_ESTIMATE',))
    rand_arr_idft = idft(rand_arr_dft, flags=('FFTW_ESTIMATE',))
    assert (rand_arr_idft - rand_arr).norm() < 1e-6

    # 2d, halfcomplex, first axis
    axes = 0
    dft_dom = _dft_complex_space.real_space
    dft = DiscreteFourierTransform(domain=dft_dom, impl=impl, halfcomplex=True,
                                   axes=axes)
    idft = DiscreteFourierTransformInverse(range=dft_dom, impl=impl,
                                           halfcomplex=True, axes=axes)

    assert dft.domain == idft.range
    assert dft.range == idft.domain

    one = dft.domain.one()
    one_dft = dft(one, flags=('FFTW_ESTIMATE',))
    true_dft = [[4, 4, 4, 4, 4],  # transform axis shortened
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0]]
    assert all_almost_equal(one_dft, true_dft)

    one_idft1 = idft(one_dft, flags=('FFTW_ESTIMATE',))
    one_idft2 = dft.inverse(one_dft, flags=('FFTW_ESTIMATE',))
    assert all_almost_equal(one_idft1, one)
    assert all_almost_equal(one_idft2, one)

    rand_arr = noise_element(dft_dom)
    rand_arr_dft = dft(rand_arr, flags=('FFTW_ESTIMATE',))
    rand_arr_idft = idft(rand_arr_dft, flags=('FFTW_ESTIMATE',))
    assert (rand_arr_idft - rand_arr).norm() < 1e-6


def test_dft_sign(impl, _dft_complex_space):
    # Test if the FT sign behaves as expected, i.e. that the FT with sign
    # '+' and '-' have same real parts and opposite imaginary parts.
    # 2d, complex, all ones and random back & forth
    skip_incompatible_impl(impl, _dft_complex_space)
    dft_dom = _dft_complex_space
    dft_minus = DiscreteFourierTransform(domain=dft_dom, impl=impl, sign='-')
    dft_plus = DiscreteFourierTransform(domain=dft_dom, impl=impl, sign='+')

    arr = dft_dom.element([[0, 0, 0, 0, 0],
                           [0, 0, 1, 1, 0],
                           [0, 0, 1, 1, 0],
                           [0, 0, 0, 0, 0]])
    arr_dft_minus = dft_minus(arr, flags=('FFTW_ESTIMATE',))
    arr_dft_plus = dft_plus(arr, flags=('FFTW_ESTIMATE',))

    assert all_almost_equal(arr_dft_minus.real, arr_dft_plus.real)
    assert all_almost_equal(arr_dft_minus.imag, -arr_dft_plus.imag)
    assert all_almost_equal(dft_minus.inverse(arr_dft_minus), arr)
    assert all_almost_equal(dft_plus.inverse(arr_dft_plus), arr)
    assert all_almost_equal(dft_minus.inverse.inverse(arr), dft_minus(arr))
    assert all_almost_equal(dft_plus.inverse.inverse(arr), dft_plus(arr))

    # 2d, halfcomplex, first axis
    axes = (0,)
    dft_dom = _dft_complex_space.real_space
    arr = dft_dom.element([[0, 0, 0, 0, 0],
                           [0, 0, 1, 1, 0],
                           [0, 0, 1, 1, 0],
                           [0, 0, 0, 0, 0]])

    dft = DiscreteFourierTransform(
        domain=dft_dom, impl=impl, halfcomplex=True, sign='-', axes=axes)

    arr_dft_minus = dft(arr, flags=('FFTW_ESTIMATE',))
    arr_idft_minus = dft.inverse(arr_dft_minus, flags=('FFTW_ESTIMATE',))

    assert all_almost_equal(arr_idft_minus, arr)

    with pytest.raises(ValueError):
        DiscreteFourierTransform(
            domain=dft_dom, impl=impl, halfcomplex=True, sign='+', axes=axes)


def test_dft_init_plan(impl):

    # 2d, halfcomplex, first axis
    shape = (4, 5)
    axes = 0
    dft_dom = _dft_space(shape, dtype='float32')

    dft = DiscreteFourierTransform(dft_dom, impl=impl, axes=axes,
                                   halfcomplex=True)

    if impl != 'pyfftw':
        with pytest.raises(ValueError):
            dft.init_fftw_plan()
        with pytest.raises(ValueError):
            dft.clear_fftw_plan()
    else:
        dft.init_fftw_plan()

        # Make sure plan can be used
        dft._fftw_plan(dft.domain.element().asarray(),
                       dft.range.element().asarray())
        dft.clear_fftw_plan()
        assert dft._fftw_plan is None


# ---- FourierTransform ---- #


def test_fourier_trafo_range(exponent, odl_floating_dtype, odl_impl_device_pairs):
    # Check if the range is initialized correctly. Encompasses the init test
    dtype = odl_floating_dtype

    # Testing R2C for real dtype, else C2C
    impl, device = odl_impl_device_pairs

    # 1D
    shape = 10
    space_discr = odl.uniform_discr(0, 1, shape, exponent=exponent,
                                    impl=impl, device=device, dtype=dtype)

    dft = FourierTransform(space_discr, halfcomplex=True, shift=True)
    assert dft.range.field == odl.ComplexNumbers()
    halfcomplex = True if is_real_dtype(dtype) else False
    assert dft.range.grid == reciprocal_grid(dft.domain.grid,
                                             halfcomplex=halfcomplex,
                                             shift=True)
    assert dft.range.exponent == conj_exponent(exponent)

    # 3D
    shape = (3, 4, 5)
    space_discr = odl.uniform_discr([0] * 3, [1] * 3, shape, exponent=exponent,
                                    impl=impl, device=device, dtype=dtype)

    dft = FourierTransform(space_discr, halfcomplex=True, shift=True)
    assert dft.range.field == odl.ComplexNumbers()
    halfcomplex = True if is_real_dtype(dtype) else False
    assert dft.range.grid == reciprocal_grid(dft.domain.grid,
                                             halfcomplex=halfcomplex,
                                             shift=True)
    assert dft.range.exponent == conj_exponent(exponent)

    # shift must be True in the last axis
    if halfcomplex:
        with pytest.raises(ValueError):
            FourierTransform(space_discr, shift=(True, True, False))

    if exponent != 2.0:
        with pytest.raises(NotImplementedError):
            dft.adjoint

    with pytest.raises(TypeError):
        FourierTransform(dft.domain.partition)


def test_fourier_trafo_init_plan(impl, odl_floating_dtype):
    dtype = odl_floating_dtype

    # Not supported, skip
    if dtype == np.dtype('float16') and impl == 'pyfftw':
        return

    shape = 10
    halfcomplex, _ = _params_from_dtype(dtype)

    space_discr = odl.uniform_discr(0, 1, shape, dtype=dtype)

    ft = FourierTransform(space_discr, impl=impl, halfcomplex=halfcomplex)
    if impl != 'pyfftw':
        with pytest.raises(ValueError):
            ft.init_fftw_plan()
        with pytest.raises(ValueError):
            ft.clear_fftw_plan()
    else:
        ft.init_fftw_plan()

        # Make sure plan can be used
        ft._fftw_plan(ft.domain.element().asarray(),
                      ft.range.element().asarray())
        ft.clear_fftw_plan()
        assert ft._fftw_plan is None

    # With temporaries
    ft.create_temporaries(r=True, f=False)
    if impl != 'pyfftw':
        with pytest.raises(ValueError):
            ft.init_fftw_plan()
        with pytest.raises(ValueError):
            ft.clear_fftw_plan()
    else:
        ft.init_fftw_plan()

        # Make sure plan can be used
        ft._fftw_plan(ft.domain.element().asarray(),
                      ft.range.element().asarray())
        ft.clear_fftw_plan()
        assert ft._fftw_plan is None

    ft.create_temporaries(r=False, f=True)
    if impl != 'pyfftw':
        with pytest.raises(ValueError):
            ft.init_fftw_plan()
        with pytest.raises(ValueError):
            ft.clear_fftw_plan()
    else:
        ft.init_fftw_plan()

        # Make sure plan can be used
        ft._fftw_plan(ft.domain.element().asarray(),
                      ft.range.element().asarray())
        ft.clear_fftw_plan()
        assert ft._fftw_plan is None


def test_fourier_trafo_create_temp():

    shape = 10
    space_discr = odl.uniform_discr(0, 1, shape, dtype='complex64')

    ft = FourierTransform(space_discr)
    ft.create_temporaries()
    assert ft._tmp_r is not None
    assert ft._tmp_f is not None

    ift = ft.inverse
    assert ift._tmp_r is not None
    assert ift._tmp_f is not None

    ft.clear_temporaries()
    assert ft._tmp_r is None
    assert ft._tmp_f is None


def test_fourier_trafo_call(impl, odl_floating_dtype, odl_impl_device_pairs):
    # Test if all variants can be called without error
    backend_impl, device = odl_impl_device_pairs
    dtype = odl_floating_dtype

    # Not supported, skip
    if dtype == np.dtype('float16') and impl == 'pyfftw':
        return

    shape = 10
    halfcomplex, _ = _params_from_dtype(dtype)
    space_discr = odl.uniform_discr(0, 1, shape, dtype=dtype, impl=backend_impl, device=device)

    skip_incompatible_impl(impl, space_discr)

    ft = FourierTransform(space_discr, impl=impl, halfcomplex=halfcomplex)
    ift = ft.inverse

    one = space_discr.one()
    assert odl.allclose(ift(ft(one)), one)

    # With temporaries
    ft.create_temporaries()
    ift = ft.inverse  # shares temporaries
    one = space_discr.one()
    assert odl.allclose(ift(ft(one)), one)


def test_fourier_trafo_charfun_1d(odl_impl_device_pairs):
    # Characteristic function of [0, 1], its Fourier transform is
    # given by exp(-1j * y / 2) * sinc(y/2)
    impl, device = odl_impl_device_pairs
    def char_interval(x):
        return (x.real >= 0) & (x.real <= 1)

    def char_interval_ft(x):
        if impl == 'numpy':
            return np.exp(-1j * x / 2) * sinc(x / 2) / np.sqrt(2 * np.pi)
        
        elif impl == 'pytorch':
            import torch
            return torch.exp(-1j * x / 2) * sinc_pytorch(x / 2) / np.sqrt(2 * np.pi)
        else:
            raise NotImplementedError

    # Base version
    discr = odl.uniform_discr(-2, 2, 40, impl=impl, device=device)
    dft_base = FourierTransform(discr)

    # Complex version, should be as good
    discr = odl.uniform_discr(-2, 2, 40, impl=impl, device=device, dtype='complex64')
    dft_complex = FourierTransform(discr)

    # Without shift
    discr = odl.uniform_discr(-2, 2, 40, impl=impl, device=device, dtype='complex64')
    dft_complex_shift = FourierTransform(discr, shift=False)

    for dft in [dft_base, dft_complex, dft_complex_shift]:
        func_true_ft = dft.range.element(char_interval_ft)
        func_dft = dft(char_interval)
        assert (func_dft - func_true_ft).norm() < 5e-6


def test_fourier_trafo_scaling(odl_impl_device_pairs):
    # Test if the FT scales correctly

    # Characteristic function of [0, 1], its Fourier transform is
    # given by exp(-1j * y / 2) * sinc(y/2)
    impl, device = odl_impl_device_pairs

    def char_interval(x):
        return (x.real >= 0) & (x.real <= 1)

    def char_interval_ft(x):
        if impl == 'numpy':
            return np.exp(-1j * x / 2) * sinc(x / 2) / np.sqrt(2 * np.pi)
        
        elif impl == 'pytorch':
            import torch
            return torch.exp(-1j * x / 2) * sinc_pytorch(x / 2) / np.sqrt(2 * np.pi)
        else:
            raise NotImplementedError

    discr = odl.uniform_discr(-2, 2, 40, impl=impl, device=device, dtype='complex128')
    dft = FourierTransform(discr)

    for factor in (2, 1j, -2.5j, 1 - 4j):
        func_true_ft = factor * dft.range.element(char_interval_ft)
        func_dft = dft(factor * discr.element(char_interval))
        assert (func_dft - func_true_ft).norm() < 1e-6


def test_fourier_trafo_sign(impl,odl_real_floating_dtype,odl_impl_device_pairs):
    # Test if the FT sign behaves as expected, i.e. that the FT with sign
    # '+' and '-' have same real parts and opposite imaginary parts.
    impl_backend, device = odl_impl_device_pairs
    discrspace_dtype = complex_dtype(odl_real_floating_dtype)

    def char_interval(x):
        return (x.real >= 0) & (x.real <= 1)

    discr = odl.uniform_discr(-2, 2, 40, impl=impl_backend, device=device, dtype=discrspace_dtype)

    skip_incompatible_impl(impl,discr)

    ft_minus = FourierTransform(discr, sign='-', impl=impl)
    ft_plus = FourierTransform(discr, sign='+', impl=impl)

    func_ft_minus = ft_minus(char_interval)
    func_ft_plus = ft_plus(char_interval)

    data = func_ft_minus.data
    if impl_backend == 'pytorch':
        data = data.detach().cpu().numpy()

    if odl_real_floating_dtype == "float16":
        tolerance = np.linalg.norm(data) * 1e-3
    elif odl_real_floating_dtype == "float32" or odl_real_floating_dtype == float:
        tolerance = np.linalg.norm(data) * 1e-7
    elif odl_real_floating_dtype == "float64" :
        tolerance = np.linalg.norm(data) * 1e-15
    elif odl_real_floating_dtype == "float128":
        if np.__version__<'2':
            # NumPy-1 does not use quadruple precision for the FFT, but double precision
            # and converts the result, so we do not achieve closer tolerance there.
            tolerance = np.linalg.norm(data) * 1e-15
        else:
            tolerance = np.linalg.norm(data) * 1e-19
    else:
        raise TypeError(f"No known tolerance for dtype {odl_real_floating_dtype}")

    def assert_close(x,y):
        x = x.data
        y = y.data
        if impl_backend == 'pytorch':
            x = x.detach().cpu().numpy()
            y = y.detach().cpu().numpy()
        assert(np.linalg.norm(x-y) < tolerance)

    assert_close(func_ft_minus.real, func_ft_plus.real)
    assert_close(func_ft_minus.imag, -func_ft_plus.imag)
    assert_close(ft_minus.inverse.inverse(char_interval), ft_minus(char_interval))
    assert_close(ft_plus.inverse.inverse(char_interval), ft_plus(char_interval))

    discr = odl.uniform_discr(-2, 2, 40, impl=impl_backend, device=device, dtype='float32')
    with pytest.raises(ValueError):
        FourierTransform(discr, sign='+', impl=impl, halfcomplex=True)
    with pytest.raises(ValueError):
        FourierTransform(discr, sign=-1, impl=impl)


def test_fourier_trafo_inverse(impl, sign, odl_impl_device_pairs):
    # Test if the inverse really is the inverse

    impl_backend, device = odl_impl_device_pairs

    def char_interval(x):
        return (x.real >= 0) & (x.real <= 1)

    # Complex-to-complex
    discr = odl.uniform_discr(-2, 2, 40, impl=impl_backend, device=device, dtype='complex64')

    skip_incompatible_impl(impl,discr)

    discr_char = discr.element(char_interval)

    ft = FourierTransform(discr, sign=sign, impl=impl)
    assert all_almost_equal(ft.inverse(ft(char_interval)), discr_char)
    assert all_almost_equal(ft.adjoint(ft(char_interval)), discr_char)

    # Half-complex
    discr = odl.uniform_discr(-2, 2, 40, impl=impl_backend, device=device,dtype='float32')
    ft = FourierTransform(discr, impl=impl, halfcomplex=True)
    assert all_almost_equal(ft.inverse(ft(char_interval)), discr_char.real)

    def char_rect(x):
        return (x[0].real >= 0) & (x[0].real <= 1) & (x[1].real >= 0) & (x[1].real <= 1)

    # 2D with axes, C2C
    discr = odl.uniform_discr([-2, -2], [2, 2], (20, 10), impl=impl_backend,
                              device=device, dtype='complex64')
    discr_rect = discr.element(char_rect)

    for axes in [(0,), 1]:
        ft = FourierTransform(discr, sign=sign, impl=impl, axes=axes)
        assert all_almost_equal(ft.inverse(ft(char_rect)), discr_rect)
        assert all_almost_equal(ft.adjoint(ft(char_rect)), discr_rect)

    # 2D with axes, halfcomplex
    discr = odl.uniform_discr([-2, -2], [2, 2], (20, 10), impl=impl_backend,
                              device=device, dtype='float32')
    discr_rect = discr.element(char_rect)

    for halfcomplex in [False, True]:
        if halfcomplex and sign == '+':
            continue  # cannot mix halfcomplex with sign

        for axes in [(0,), (1,)]:
            ft = FourierTransform(discr, sign=sign, impl=impl, axes=axes,
                                  halfcomplex=halfcomplex)
            assert all_almost_equal(ft.inverse(ft(char_rect)), discr_rect)
            assert all_almost_equal(ft.adjoint(ft(char_rect)), discr_rect)


def test_fourier_trafo_hat_1d(odl_impl_device_pairs):
    # Hat function as used in linear interpolation. It is not so
    # well discretized by nearest neighbor interpolation, so a larger
    # error is to be expected.
    impl, device = odl_impl_device_pairs
    def hat_func(x):
        if impl == 'numpy':
            out = np.where(x < 0, 1 + x, 1 - x)
            out[x < -1] = 0
            out[x > 1] = 0
            return out
        elif impl == 'pytorch':
            import torch
            out = torch.where(x < 0, 1 + x, 1 - x)
            out[x < -1] = 0
            out[x > 1] = 0
            return out
        else:
            raise NotImplementedError

    def hat_func_ft(x):
        if impl == 'numpy':
            return sinc(x / 2) ** 2 / np.sqrt(2 * np.pi)
        elif impl == 'pytorch':
            return sinc_pytorch(x / 2) ** 2 / np.sqrt(2 * np.pi)
        else:
            raise NotImplementedError
    # Using a single-precision implementation, should be as good
    discr = odl.uniform_discr(-2, 2, 101, impl=impl, device=device, dtype='float32')
    dft = FourierTransform(discr)
    func_true_ft = dft.range.element(hat_func_ft)
    func_dft = dft(hat_func)
    assert (func_dft - func_true_ft).norm() < 0.001


def test_fourier_trafo_complex_sum(odl_impl_device_pairs):
    # Sum of characteristic function and hat function, both with
    # known FT's.

    impl, device = odl_impl_device_pairs

    def hat_func(x):
        if impl == 'numpy':
            out = np.where(x < 0, 1 + x, 1 - x)
            out[x < -1] = 0
            out[x > 1] = 0
            return out
        elif impl == 'pytorch':
            import torch
            out = torch.where(x.real < 0, 1 + x.real, 1 - x.real)
            out[x.real < -1] = 0
            out[x.real > 1] = 0
            return out
        else:
            raise NotImplementedError

    def hat_func_ft(x):
        if impl == 'numpy':
            return sinc(x / 2) ** 2 / np.sqrt(2 * np.pi)
        elif impl == 'pytorch':
            return sinc_pytorch(x / 2) ** 2 / np.sqrt(2 * np.pi)
        else:
            raise NotImplementedError

    def char_interval(x):
        return (x.real >= 0) & (x.real <= 1)

    def char_interval_ft(x):
        if impl == 'numpy':
            return np.exp(-1j * x / 2) * sinc(x / 2) / np.sqrt(2 * np.pi)
        
        elif impl == 'pytorch':
            import torch
            return torch.exp(-1j * x / 2) * sinc_pytorch(x / 2) / np.sqrt(2 * np.pi)
        else:
            raise NotImplementedError

    discr = odl.uniform_discr(-2, 2, 200, impl=impl, device=device, dtype='complex128')
    dft = FourierTransform(discr, shift=False)

    func = discr.element(hat_func) + 1j * discr.element(char_interval)
    func_true_ft = (
        dft.range.element(hat_func_ft)
        + 1j * dft.range.element(char_interval_ft)
    )
    func_dft = dft(func)
    assert (func_dft - func_true_ft).norm() < 0.001


def test_fourier_trafo_gaussian_1d(odl_impl_device_pairs):
    # Gaussian function, will be mapped to itself. Truncation error is
    # relatively large, though, we need a large support.

    impl, device = odl_impl_device_pairs

    def gaussian(x):
        if impl == 'numpy':
            return np.exp(-x ** 2 / 2)
        elif impl == 'pytorch':
            import torch
            return torch.exp(-x ** 2 / 2)
        else:
            raise NotImplementedError
    discr = odl.uniform_discr(-10, 10, 201, impl=impl, device=device)
    dft = FourierTransform(discr)
    func_true_ft = dft.range.element(gaussian)
    func_dft = dft(gaussian)
    assert (func_dft - func_true_ft).norm() < 0.001


def test_fourier_trafo_freq_shifted_charfun_1d(odl_impl_device_pairs):
    # Frequency-shifted characteristic function: mult. with
    # exp(-1j * b * x) corresponds to shifting the FT by b.
    impl, device=odl_impl_device_pairs
    
    def fshift_char_interval(x):
        if impl == 'numpy':
            return np.exp(-1j * x * np.pi) * ((x >= -0.5) & (x <= 0.5))
        elif impl== 'pytorch':
             import torch
             return torch.exp(-1j * x * np.pi) * ((x.real >= -0.5) & (x.real <= 0.5))
        else:
            raise NotImplementedError

    def fshift_char_interval_ft(x):
        if impl=='numpy':    
            return sinc((x + np.pi) / 2) / np.sqrt(2 * np.pi)
        elif impl== 'pytorch':
            return sinc_pytorch((x + np.pi) / 2) / np.sqrt(2 * np.pi)
        else:
            raise NotImplementedError
    # Number of points is very important here (aliasing)
    discr = odl.uniform_discr(-2, 2, 400, impl=impl, device=device,
                              dtype='complex64')
    dft = FourierTransform(discr)
    func_true_ft = dft.range.element(fshift_char_interval_ft)
    func_dft = dft(fshift_char_interval)
    assert (func_dft - func_true_ft).norm() < 0.001


def test_dft_with_known_pairs_2d(odl_impl_device_pairs):
    impl, device=odl_impl_device_pairs
    # Frequency-shifted product of characteristic functions
    def fshift_char_rect(x):
        # Characteristic function of the cuboid
        # [-1, 1] x [1, 2]
        if impl =='numpy':
            return (x[0] >= -1) & (x[0] <= 1) & (x[1] >= 1) & (x[1] <= 2)
        elif impl == 'pytorch':
            return (x[0].real >= -1) & (x[0].real <= 1) & (x[1].real >= 1) & (x[1].real <= 2)
        else:
            raise NotImplementedError

    def fshift_char_rect_ft(x):
        # FT is a product of shifted and frequency-shifted sinc functions
        # 1st comp.: 2 * sinc(y)
        # 2nd comp.: exp(-1j * y * 3/2) * sinc(y/2)
        # Overall factor: (2 * pi)^(-1)
        if impl =='numpy':
            return (
                2 * sinc(x[0])
                * np.exp(-1j * x[1] * 3 / 2) * sinc(x[1] / 2)
                / (2 * np.pi)
            )
        elif impl == 'pytorch':
            import torch
            return (
                2 * sinc_pytorch(x[0])
                * torch.exp(-1j * x[1] * 3 / 2) * sinc_pytorch(x[1] / 2)
                / (2 * np.pi)
            )

        else:
            raise NotImplementedError

    discr = odl.uniform_discr([-2] * 2, [2] * 2, (100, 400), impl=impl, device=device, dtype='complex64')
    dft = FourierTransform(discr)
    func_true_ft = dft.range.element(fshift_char_rect_ft)
    func_dft = dft(fshift_char_rect)
    assert (func_dft - func_true_ft).norm() < 0.001


def test_fourier_trafo_completely(odl_impl_device_pairs):
    # Complete explicit test of all FT components on two small examples
    impl, device = odl_impl_device_pairs
    
    # Discretization with 4 points
    discr = odl.uniform_discr(-2, 2, 4, impl=impl, device=device, dtype=complex)
    # Interval boundaries -2, -1, 0, 1, 2
    assert np.allclose(discr.partition.cell_boundary_vecs[0],
                       [-2, -1, 0, 1, 2])
    # Grid points -1.5, -0.5, 0.5, 1.5
    assert np.allclose(discr.grid.coord_vectors[0],
                       [-1.5, -0.5, 0.5, 1.5])

    # First test function, symmetric. Can be represented exactly in the
    # discretization.
    def f(x):
        if impl == 'numpy':
            return (x >= -1) & (x <= 1)
        elif impl == 'pytorch':
            return (x.real >= -1) & (x.real <= 1)
        else:
            raise NotImplementedError


    def fhat(x):
        if impl == 'numpy':
            return np.sqrt(2 / np.pi) * sinc(x)
        elif impl == 'pytorch':
            return  np.sqrt(2 / np.pi) * sinc_pytorch(x)
        else:
            raise NotImplementedError

    # Discretize f, check values
    f_discr = discr.element(f)
    assert all_almost_equal(f_discr, [0, 1, 1, 0])

    # "s" = shifted, "n" = not shifted

    # Reciprocal grids
    recip_s = reciprocal_grid(discr.grid, shift=True)
    recip_n = reciprocal_grid(discr.grid, shift=False)
    assert np.allclose(recip_s.coord_vectors[0],
                       np.linspace(-np.pi, np.pi / 2, 4))
    assert np.allclose(recip_n.coord_vectors[0],
                       np.linspace(-3 * np.pi / 4, 3 * np.pi / 4, 4))

    # Range
    range_part_s = odl.uniform_partition_fromgrid(recip_s)
    range_s = odl.uniform_discr_frompartition(range_part_s, dtype=complex, impl=impl, device=device)
    range_part_n = odl.uniform_partition_fromgrid(recip_n)
    range_n = odl.uniform_discr_frompartition(range_part_n, dtype=complex, impl=impl, device=device)

    namespace = discr.array_namespace
    backend = discr.array_backend

    # Pre-processing
    preproc_s = [1, -1, 1, -1]
    preproc_n = [np.exp(1j * 3 / 4 * np.pi * k) for k in range(4)]

    fpre_s = dft_preprocess_data(f_discr, shift=True)
    fpre_n = dft_preprocess_data(f_discr, shift=False)
    assert all_almost_equal(fpre_s, f_discr * discr.element(preproc_s))
    assert all_almost_equal(fpre_n, f_discr * discr.element(preproc_n))

    # FFT step, replicating the _call_numpy method
    fft_s = namespace.fft.fftn(fpre_s, s=discr.shape, axes=[0])
    fft_n = namespace.fft.fftn(fpre_n, s=discr.shape, axes=[0])
    s = backend.array_constructor([0, -1 + 1j, 2, -1 - 1j], device=device, dtype=complex)
    n = backend.array_constructor([
            np.exp(1j * np.pi * (3 - 2 * k) / 4)
            + np.exp(1j * np.pi * (3 - 2 * k) / 2)
            for k in range(4)], device=device, dtype=complex
         )
    assert namespace.allclose(fft_s, s)
    assert namespace.allclose(fft_n, n)

    # Interpolation kernel FT
    interp_s = np.sinc(np.linspace(-1 / 2, 1 / 4, 4)) / np.sqrt(2 * np.pi)
    interp_n = np.sinc(np.linspace(-3 / 8, 3 / 8, 4)) / np.sqrt(2 * np.pi)
    assert np.allclose(interp_s,
                       _interp_kernel_ft(np.linspace(-1 / 2, 1 / 4, 4),
                                         interp='nearest'))
    assert np.allclose(interp_n,
                       _interp_kernel_ft(np.linspace(-3 / 8, 3 / 8, 4),
                                         interp='nearest'))

    # Post-processing
    postproc_s = np.exp(1j * np.pi * np.linspace(-3 / 2, 3 / 4, 4))
    postproc_n = np.exp(1j * np.pi * np.linspace(-9 / 8, 9 / 8, 4))

    fpost_s = dft_postprocess_data(
        range_s.element(fft_s), real_grid=discr.grid, recip_grid=recip_s,
        shift=[True], axes=(0,), interp='nearest')
    fpost_n = dft_postprocess_data(
        range_n.element(fft_n), real_grid=discr.grid, recip_grid=recip_n,
        shift=[False], axes=(0,), interp='nearest')
    
    postproc_s = backend.array_constructor(postproc_s, device=device)
    postproc_n = backend.array_constructor(postproc_n, device=device)

    interp_s = backend.array_constructor(interp_s, device=device)
    interp_n = backend.array_constructor(interp_n, device=device)

    assert namespace.allclose(fpost_s, fft_s * postproc_s * interp_s)
    assert namespace.allclose(fpost_n, fft_n * postproc_n * interp_n)

    # Comparing to the known result sqrt(2/pi) * sinc(x)
    recip_s_array = backend.array_constructor(recip_s.coord_vectors[0], device=device)
    recip_n_array = backend.array_constructor(recip_n.coord_vectors[0], device=device)
    assert namespace.allclose(fpost_s.real, fhat(recip_s_array))
    assert namespace.allclose(fpost_n.real, fhat(recip_n_array))

    # Doing the exact same with direct application of the FT operator
    ft_op_s = FourierTransform(discr, shift=True)
    ft_op_n = FourierTransform(discr, shift=False)
    assert ft_op_s.range.grid == recip_s
    assert ft_op_n.range.grid == recip_n

    ft_f_s = ft_op_s(f)
    ft_f_n = ft_op_n(f)
    assert all_almost_equal(ft_f_s.real, fhat(recip_s_array))
    assert all_almost_equal(ft_f_n.real, fhat(recip_n_array))

    # Second test function, asymmetric. Can also be represented exactly in the
    # discretization.
    def f(x):
        return (x.real >= 0) & (x.real <= 1)

    def fhat(x):
        if impl == 'numpy':
            return np.exp(-1j * x / 2) * sinc(x / 2) / np.sqrt(2 * np.pi)
        elif impl == 'pytorch':
            return namespace.exp(-1j * x / 2) * sinc_pytorch(x / 2) / np.sqrt(2 * np.pi)
        else:
            raise NotImplementedError
    # Discretize f, check values
    f_discr = discr.element(f)
    assert all_almost_equal(f_discr, [0, 0, 1, 0])

    # Pre-processing
    fpre_s = dft_preprocess_data(f_discr, shift=True)
    fpre_n = dft_preprocess_data(f_discr, shift=False)
    assert all_almost_equal(fpre_s, [0, 0, 1, 0])
    assert all_almost_equal(fpre_n, backend.array_constructor([0, 0, -1j, 0], device=device, dtype=complex))

    # FFT step
    fft_s = namespace.fft.fftn(fpre_s, s=discr.shape, axes=[0])
    fft_n = namespace.fft.fftn(fpre_n, s=discr.shape, axes=[0])
    assert all_almost_equal(fft_s, [1, -1, 1, -1])
    assert all_almost_equal(fft_n, backend.array_constructor([-1j, 1j, -1j, 1j], device=device, dtype=complex))

    fpost_s = dft_postprocess_data(
        range_s.element(fft_s), real_grid=discr.grid, recip_grid=recip_s,
        shift=[True], axes=(0,), interp='nearest')
    fpost_n = dft_postprocess_data(
        range_n.element(fft_n), real_grid=discr.grid, recip_grid=recip_n,
        shift=[False], axes=(0,), interp='nearest')

    assert all_almost_equal(fpost_s, fft_s * postproc_s * interp_s)
    assert all_almost_equal(fpost_n, fft_n * postproc_n * interp_n)

    # Comparing to the known result exp(-1j*x/2) * sinc(x/2) / sqrt(2*pi)
    recip_s = backend.array_constructor(recip_s.coord_vectors[0], device=device)
    recip_n = backend.array_constructor(recip_n.coord_vectors[0], device=device)
    assert all_almost_equal(fpost_s, fhat(recip_s))
    assert all_almost_equal(fpost_n, fhat(recip_n))

    # Doing the exact same with direct application of the FT operator
    ft_f_s = ft_op_s(f)
    ft_f_n = ft_op_n(f)
    assert all_almost_equal(ft_f_s, fhat(recip_s))
    assert all_almost_equal(ft_f_n, fhat(recip_n))


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Test reconstruction with ASTRA."""

from __future__ import division

import numpy as np
import pytest
from packaging.version import parse as parse_version

import odl
from odl.applications.tomo.util.testutils import (
    skip_if_no_astra, skip_if_no_astra_cuda, skip_if_no_skimage)
from odl.core.util.dtype_utils import _universal_dtype_identifier
from odl.core.util.testutils import all_almost_equal, simple_fixture

# --- pytest fixtures --- #


# Global pytest mark
pytestmark = pytest.mark.suite('largescale')


dtype_params = ['float32', 'complex64']
dtype = simple_fixture('dtype', dtype_params)


# Find the valid projectors
projectors = []
projectors.extend(
    (pytest.param(proj_cfg, marks=skip_if_no_astra)
     for proj_cfg in ['par2d astra_cpu uniform',
                      'par2d astra_cpu nonuniform',
                      'par2d astra_cpu random',
                      'cone2d astra_cpu uniform',
                      'cone2d astra_cpu nonuniform',
                      'cone2d astra_cpu random'])
)
projectors.extend(
    (pytest.param(proj_cfg, marks=skip_if_no_astra_cuda)
     for proj_cfg in ['par2d astra_cuda uniform',
                      'par2d astra_cuda nonuniform',
                      'par2d astra_cuda random',
                      'cone2d astra_cuda uniform',
                      'cone2d astra_cuda nonuniform',
                      'cone2d astra_cuda random',
                      'par3d astra_cuda uniform',
                      'par3d astra_cuda nonuniform',
                      'par3d astra_cuda random',
                      'cone3d astra_cuda uniform',
                      'cone3d astra_cuda nonuniform',
                      'cone3d astra_cuda random',
                      'helical astra_cuda uniform'])
)
projectors.extend(
    (pytest.param(proj_cfg, marks=skip_if_no_skimage)
     for proj_cfg in ['par2d skimage uniform'])
)

projector_ids = [
    " geom='{}' - impl='{}' - angles='{}' ".format(*p.values[0].split())
    for p in projectors
]


weighting = simple_fixture('weighting', [None, 1.0])


@pytest.fixture(scope="module", params=projectors, ids=projector_ids)
def projector(request, dtype, weighting, odl_impl_device_pairs):
    array_impl, device = odl_impl_device_pairs
    print(f"{array_impl=}, {device=}")

    n_angles = 200

    geom, ray_impl, angles = request.param.split()

    if angles == 'uniform':
        apart = odl.uniform_partition(0, 2 * np.pi, n_angles)
    elif angles == 'random':
        # Linearly spaced with random noise
        min_pt = 2 * (2.0 * np.pi) / n_angles
        max_pt = (2.0 * np.pi) - 2 * (2.0 * np.pi) / n_angles
        points = np.linspace(min_pt, max_pt, n_angles)
        points += np.random.rand(n_angles) * (max_pt - min_pt) / (5 * n_angles)
        apart = odl.nonuniform_partition(points)
    elif angles == 'nonuniform':
        # Angles spaced quadratically
        min_pt = 2 * (2.0 * np.pi) / n_angles
        max_pt = (2.0 * np.pi) - 2 * (2.0 * np.pi) / n_angles
        points = np.linspace(min_pt ** 0.5, max_pt ** 0.5, n_angles) ** 2
        apart = odl.nonuniform_partition(points)
    else:
        assert False

    if geom == 'par2d':
        # Reconstruction space
        reco_space = odl.uniform_discr([-20, -20], [20, 20], [100, 100],
                                       dtype=dtype, weighting=weighting, impl=array_impl, device=device)

        # Geometry
        dpart = odl.uniform_partition(-30, 30, 200)
        geom = odl.applications.tomo.Parallel2dGeometry(apart, dpart)

        # Ray transform
        return odl.applications.tomo.RayTransform(reco_space, geom, impl=ray_impl)

    elif geom == 'par3d':
        # Reconstruction space
        reco_space = odl.uniform_discr([-20, -20, -20], [20, 20, 20],
                                       [100, 100, 100],
                                       dtype=dtype, weighting=weighting, impl=array_impl, device=device)

        # Geometry
        dpart = odl.uniform_partition([-30, -30], [30, 30], [200, 200])
        geom = odl.applications.tomo.Parallel3dAxisGeometry(apart, dpart, axis=[1, 0, 0])

        # Ray transform
        return odl.applications.tomo.RayTransform(reco_space, geom, impl=ray_impl)

    elif geom == 'cone2d':
        # Reconstruction space
        reco_space = odl.uniform_discr([-20, -20], [20, 20], [100, 100],
                                       dtype=dtype, impl=array_impl, device=device)

        # Geometry
        dpart = odl.uniform_partition(-30, 30, 200)
        geom = odl.applications.tomo.FanBeamGeometry(apart, dpart, src_radius=200,
                                        det_radius=100)

        # Ray transform
        return odl.applications.tomo.RayTransform(reco_space, geom, impl=ray_impl)

    elif geom == 'cone3d':
        # Reconstruction space
        reco_space = odl.uniform_discr([-20, -20, -20], [20, 20, 20],
                                       [100, 100, 100], dtype=dtype, impl=array_impl, device=device)

        # Geometry
        dpart = odl.uniform_partition([-30, -30], [30, 30], [200, 200])
        geom = odl.applications.tomo.ConeBeamGeometry(
            apart, dpart, src_radius=200, det_radius=100, axis=[1, 0, 0])

        # Ray transform
        return odl.applications.tomo.RayTransform(reco_space, geom, impl=ray_impl)

    elif geom == 'helical':
        # Reconstruction space
        reco_space = odl.uniform_discr([-20, -20, 0], [20, 20, 40],
                                       [100, 100, 100], dtype=dtype, impl=array_impl, device=device)

        # Geometry
        # TODO: angles
        n_angles = 700
        apart = odl.uniform_partition(0, 8 * 2 * np.pi, n_angles)
        dpart = odl.uniform_partition([-30, -3], [30, 3], [200, 20])
        geom = odl.applications.tomo.ConeBeamGeometry(apart, dpart, pitch=5.0,
                                         src_radius=200, det_radius=100)

        # Ray transform
        return odl.applications.tomo.RayTransform(reco_space, geom, impl=ray_impl)
    else:
        raise ValueError('param not valid')


# --- RayTransform tests --- #


def test_adjoint(projector):
    """Test RayTransform adjoint matches definition."""
    # Relative tolerance, still rather high due to imperfectly matched
    # adjoint in the cone beam case
    if (
        parse_version(odl.applications.tomo.backends.ASTRA_VERSION) < parse_version('1.8rc1')
        and isinstance(projector.geometry, odl.applications.tomo.ConeBeamGeometry)
    ):
        rtol = 0.1
    else:
        rtol = 0.05

    # Create Shepp-Logan phantom
    vol = odl.core.phantom.shepp_logan(projector.domain, modified=True)

    # Calculate projection
    proj = projector(vol)
    backproj = projector.adjoint(proj)

    # Verify the identity <Ax, Ax> = <A^* A x, x>
    result_AxAx = proj.inner(proj)
    result_xAtAx = backproj.inner(vol)
    assert result_AxAx == pytest.approx(result_xAtAx, rel=rtol)


def test_adjoint_of_adjoint(projector):
    """Test RayTransform adjoint of adjoint."""

    # Create Shepp-Logan phantom
    vol = odl.core.phantom.shepp_logan(projector.domain, modified=True)

    # Calculate projection
    proj = projector(vol)
    proj_adj_adj = projector.adjoint.adjoint(vol)

    # Verify A(x) == (A^*)^*(x)
    assert proj == proj_adj_adj

    # Calculate adjoints
    proj_adj = projector.adjoint(proj)
    proj_adj_adj_adj = projector.adjoint.adjoint.adjoint(proj)

    # Verify A^*(y) == ((A^*)^*)^*(x)
    assert all_almost_equal(proj_adj, proj_adj_adj_adj)


def test_reconstruction(projector):
    """Test RayTransform for reconstruction."""
    if (
        isinstance(projector.geometry, odl.applications.tomo.ConeBeamGeometry)
        and projector.geometry.pitch != 0
    ):
        pytest.skip('reconstruction with CG is hopeless with so few angles')

    # Create Shepp-Logan phantom
    vol = odl.core.phantom.shepp_logan(projector.domain, modified=True)

    # Project data
    projections = projector(vol)

    # Reconstruct using ODL
    recon = projector.domain.zero()
    odl.solvers.conjugate_gradient_normal(projector, recon, projections,
                                          niter=20)

    # Make sure the result is somewhat close to the actual result
    maxerr = vol.norm() * 0.5
    if np.issubdtype(_universal_dtype_identifier(projector.domain.dtype), np.complexfloating):
        # Error has double the amount of components practically
        maxerr *= np.sqrt(2)
    assert recon.dist(vol) < maxerr


if __name__ == '__main__':
    odl.core.util.test_file(__file__, ['-S', 'largescale'])



# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Alternating Direction method of Multipliers (ADMM) method variants."""

from __future__ import division
from builtins import range

from odl.core.operator import Operator, OpDomainError


__all__ = ('admm_linearized',)


def admm_linearized(x, f, g, L, tau, sigma, niter, **kwargs):
    r"""Generic linearized ADMM method for convex problems.

    ADMM stands for "Alternating Direction Method of Multipliers" and
    is a popular convex optimization method. This variant solves problems
    of the form ::

        min_x [ f(x) + g(Lx) ]

    with convex ``f`` and ``g``, and a linear operator ``L``. See Section
    4.4 of `[PB2014] <http://web.stanford.edu/~boyd/papers/prox_algs.html>`_
    and the Notes for more mathematical details.

    Parameters
    ----------
    x : ``L.domain`` element
        Starting point of the iteration, updated in-place.
    f, g : `Functional`
        The functions ``f`` and ``g`` in the problem definition. They
        need to implement the ``proximal`` method.
    L : linear `Operator`
        The linear operator that is composed with ``g`` in the problem
        definition. It must fulfill ``L.domain == f.domain`` and
        ``L.range == g.domain``.
    tau, sigma : positive float
        Step size parameters for the update of the variables.
    niter : non-negative int
        Number of iterations.

    Other Parameters
    ----------------
    callback : callable, optional
        Function called with the current iterate after each iteration.

    Notes
    -----
    Given :math:`x^{(0)}` (the provided ``x``) and
    :math:`u^{(0)} = z^{(0)} = 0`, linearized ADMM applies the following
    iteration:

    .. math::
        x^{(k+1)} &= \mathrm{prox}_{\tau f} \left[
            x^{(k)} - \sigma^{-1}\tau L^*\big(
                L x^{(k)} - z^{(k)} + u^{(k)}
            \big)
        \right]

        z^{(k+1)} &= \mathrm{prox}_{\sigma g}\left(
            L x^{(k+1)} + u^{(k)}
        \right)

        u^{(k+1)} &= u^{(k)} + L x^{(k+1)} - z^{(k+1)}

    The step size parameters :math:`\tau` and :math:`\sigma` must satisfy

    .. math::
        0 < \tau < \frac{\sigma}{\|L\|^2}

    to guarantee convergence.

    The name "linearized ADMM" comes from the fact that in the
    minimization subproblem for the :math:`x` variable, this variant
    uses a linearization of a quadratic term in the augmented Lagrangian
    of the generic ADMM, in order to make the step expressible with
    the proximal operator of :math:`f`.

    Another name for this algorithm is *split inexact Uzawa method*.

    References
    ----------
    [PB2014] Parikh, N and Boyd, S. *Proximal Algorithms*. Foundations and
    Trends in Optimization, 1(3) (2014), pp 123-231.
    """
    if not isinstance(L, Operator):
        raise TypeError('`op` {!r} is not an `Operator` instance'
                        ''.format(L))

    if x not in L.domain:
        raise OpDomainError('`x` {!r} is not in the domain of `op` {!r}'
                            ''.format(x, L.domain))

    tau, tau_in = float(tau), tau
    if tau <= 0:
        raise ValueError('`tau` must be positive, got {}'.format(tau_in))

    sigma, sigma_in = float(sigma), sigma
    if sigma <= 0:
        raise ValueError('`sigma` must be positive, got {}'.format(sigma_in))

    niter, niter_in = int(niter), niter
    if niter < 0 or niter != niter_in:
        raise ValueError('`niter` must be a non-negative integer, got {}'
                         ''.format(niter_in))

    # Callback object
    callback = kwargs.pop('callback', None)
    if callback is not None and not callable(callback):
        raise TypeError('`callback` {} is not callable'.format(callback))

    # Initialize range variables
    z = L.range.zero()
    u = L.range.zero()

    # Temporary for Lx + u [- z]
    tmp_ran = L(x)
    # Temporary for L^*(Lx + u - z)
    tmp_dom = L.domain.element()

    # Store proximals since their initialization may involve computation
    prox_tau_f = f.proximal(tau)
    prox_sigma_g = g.proximal(sigma)

    for _ in range(niter):
        # tmp_ran has value Lx^k here
        # tmp_dom <- L^*(Lx^k + u^k - z^k)
        tmp_ran += u
        tmp_ran -= z
        L.adjoint(tmp_ran, out=tmp_dom)

        # x <- x^k - (tau/sigma) L^*(Lx^k + u^k - z^k)
        x.lincomb(1, x, -tau / sigma, tmp_dom)
        # x^(k+1) <- prox[tau*f](x)
        prox_tau_f(x, out=x)

        # tmp_ran <- Lx^(k+1)
        L(x, out=tmp_ran)
        # z^(k+1) <- prox[sigma*g](Lx^(k+1) + u^k)
        prox_sigma_g(tmp_ran + u, out=z)  # 1 copy here

        # u^(k+1) = u^k + Lx^(k+1) - z^(k+1)
        u += tmp_ran
        u -= z

        if callback is not None:
            callback(x)


def admm_linearized_simple(x, f, g, L, tau, sigma, niter, **kwargs):
    """Non-optimized version of ``admm_linearized``.

    This function is intended for debugging. It makes a lot of copies and
    performs no error checking.
    """
    callback = kwargs.pop('callback', None)
    z = L.range.zero()
    u = L.range.zero()
    for _ in range(niter):
        x[:] = f.proximal(tau)(x - tau / sigma * L.adjoint(L(x) + u - z))
        z = g.proximal(sigma)(L(x) + u)
        u = L(x) + u - z
        if callback is not None:
            callback(x)




"""
Sparse matrix representation for creating product space operators.
"""

import numpy as np

__all__ = ('COOMatrix',)


class COOMatrix():
    """
    Custom coo matrix representation for creating product space operators.

    The columns, rows and data are stored in separate lists such that
    A[i[k], j[k]] = data[k].

    Note that, the class is only used as a container and does not provide
    any matrix operations. Further, no checks are performed on the data thus
    duplicate and out-of-order indices are allowed and the user is responsible
    for ensuring the correct shape of the matrix.

    """

    def __init__(self, data, index, shape):

        # type check
        if len(data) != len(index[0]) or len(data) != len(index[1]):
            raise ValueError('data and index must have the same length')

        self.__data = data
        self.__row_index = np.asarray(index[0])
        self.__col_index = np.asarray(index[1])
        self.__shape = shape

    @property
    def row(self):
        """Return the row indices of the matrix."""
        return self.__row_index

    @property
    def col(self):
        """Return the column indices of the matrix."""
        return self.__col_index

    @property
    def shape(self):
        """Return the shape of the matrix."""
        return self.__shape

    @property
    def data(self):
        """Return the data of the matrix."""
        return self.__data

    def __repr__(self):
        return ( f"COO matrix({self.data},"
                + f"({self.__row_index}, {self.__col_index}), {self.shape})" )



# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Unit tests for ADMM."""

from __future__ import division
import odl
from odl.solvers import admm_linearized, Callback

from odl.core.util.testutils import all_almost_equal, noise_element


def test_admm_lin_input_handling():
    """Test to see that input is handled correctly."""

    space = odl.uniform_discr(0, 1, 10)

    L = odl.ZeroOperator(space)
    f = g = odl.functionals.ZeroFunctional(space)

    # Check that the algorithm runs. With the above operators and functionals,
    # the algorithm should not modify the initial value.
    x0 = noise_element(space)
    x = x0.copy()
    niter = 3

    admm_linearized(x, f, g, L, tau=1.0, sigma=1.0, niter=niter)

    assert x == x0

    # Check that a provided callback is actually called
    class CallbackTest(Callback):
        was_called = False

        def __call__(self, *args, **kwargs):
            self.was_called = True

    callback = CallbackTest()
    assert not callback.was_called
    admm_linearized(x, f, g, L, tau=1.0, sigma=1.0, niter=niter,
                    callback=callback)
    assert callback.was_called


def test_admm_lin_l1():
    """Verify that the correct value is returned for l1 dist optimization.

    Solves the optimization problem

        min_x ||x - data_1||_1 + 0.5 ||x - data_2||_1

    which has optimum value data_1 since the first term dominates.
    """
    space = odl.rn(5)

    L = odl.IdentityOperator(space)

    data_1 = odl.core.util.testutils.noise_element(space)
    data_2 = odl.core.util.testutils.noise_element(space)

    f = odl.functionals.L1Norm(space).translated(data_1)
    g = 0.5 * odl.functionals.L1Norm(space).translated(data_2)

    x = space.zero()
    admm_linearized(x, f, g, L, tau=1.0, sigma=2.0, niter=10)

    assert all_almost_equal(x, data_1, ndigits=2)


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import division
import pytest

import odl
from odl.core.util.testutils import all_almost_equal, simple_fixture


# base_op = simple_fixture(
#     'base_op',
#     [odl.IdentityOperator(odl.rn(3)),
#      odl.BroadcastOperator(odl.IdentityOperator(odl.rn(3)), 2),
#      odl.ReductionOperator(odl.IdentityOperator(odl.rn(3)), 2),
#      odl.DiagonalOperator(odl.IdentityOperator(odl.rn(3)), 2),
#      ],
#     fmt=' {name}={value.__class__.__name__}')

op_name = simple_fixture(name='op_name', params=['Identity', 'Broadcast', 'Reduction', 'Diagonal'])

@pytest.fixture(scope="module", )
def base_op(odl_impl_device_pairs, op_name):
    impl, device = odl_impl_device_pairs
    space = odl.rn(3, impl=impl, device=device)
    if op_name == 'Identity':
        op = odl.IdentityOperator(space)
    elif op_name == 'Broadcast':
        op = odl.BroadcastOperator(odl.IdentityOperator(space), 2)
    elif op_name == 'Reduction':
        op = odl.ReductionOperator(odl.IdentityOperator(space), 2)
    elif op_name == 'Diagonal':
        op = odl.DiagonalOperator(odl.IdentityOperator(space), 2)
    return op

def test_pspace_op_init(base_op):
    """Test initialization with different base operators."""
    A = base_op

    op = odl.ProductSpaceOperator([[A]])
    assert op.domain == A.domain ** 1
    assert op.range == A.range ** 1

    op = odl.ProductSpaceOperator([[A, A]])
    assert op.domain == A.domain ** 2
    assert op.range == A.range ** 1

    op = odl.ProductSpaceOperator([[A],
                                   [A]])
    assert op.domain == A.domain ** 1
    assert op.range == A.range ** 2

    op = odl.ProductSpaceOperator([[A, 0],
                                   [0, A]])
    assert op.domain == A.domain ** 2
    assert op.range == A.range ** 2

    op = odl.ProductSpaceOperator([[A, None],
                                   [None, A]])
    assert op.domain == A.domain ** 2
    assert op.range == A.range ** 2


def test_pspace_op_derivative(base_op):
    """Test derivatives with different base operators."""
    A = base_op

    op = odl.ProductSpaceOperator([[A + 1]])
    true_deriv = odl.ProductSpaceOperator([[A]])
    deriv = op.derivative(op.domain.zero())
    assert deriv.domain == op.domain
    assert deriv.range == op.range
    x = op.domain.one()
    assert all_almost_equal(deriv(x), true_deriv(x))

    op = odl.ProductSpaceOperator([[A + 1, 2 * A - 1]])
    deriv = op.derivative(op.domain.zero())
    true_deriv = odl.ProductSpaceOperator([[A, 2 * A]])
    assert deriv.domain == op.domain
    assert deriv.range == op.range
    x = op.domain.one()
    assert all_almost_equal(deriv(x), true_deriv(x))


def test_pspace_op_adjoint(base_op):
    """Test adjoints with different base operators."""
    A = base_op

    op = odl.ProductSpaceOperator([[A]])
    true_adj = odl.ProductSpaceOperator([[A.adjoint]])
    adj = op.adjoint
    assert adj.domain == op.range
    assert adj.range == op.domain
    y = op.range.one()
    assert all_almost_equal(adj(y), true_adj(y))

    op = odl.ProductSpaceOperator([[2 * A, -A]])
    true_adj = odl.ProductSpaceOperator([[2 * A.adjoint],
                                         [-A.adjoint]])
    adj = op.adjoint
    assert adj.domain == op.range
    assert adj.range == op.domain
    y = op.range.one()
    assert all_almost_equal(adj(y), true_adj(y))


def test_pspace_op_weighted_init(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    r3=odl.rn(3, impl=impl, device=device)
    ran = odl.ProductSpace(r3, 2, weighting=[1, 2])
    A = odl.IdentityOperator(r3)

    with pytest.raises(NotImplementedError):
        odl.ProductSpaceOperator([[A],
                                  [0]], range=ran)


def test_pspace_op_sum_call(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    r3=odl.rn(3, impl=impl, device=device)
    A = odl.IdentityOperator(r3)
    op = odl.ProductSpaceOperator([[A, A]])

    x = r3.element([1, 2, 3])
    y = r3.element([7, 8, 9])
    z = op.domain.element([x, y])

    assert all_almost_equal(op(z)[0], x + y)
    assert all_almost_equal(op(z, out=op.range.element())[0], x + y)


def test_pspace_op_project_call(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    r3=odl.rn(3, impl=impl, device=device)
    A = odl.IdentityOperator(r3)
    op = odl.ProductSpaceOperator([[A],
                                   [A]])

    x = r3.element([1, 2, 3])
    z = op.domain.element([x])

    assert x == op(z)[0]
    assert x == op(z, out=op.range.element())[0]
    assert x == op(z)[1]
    assert x == op(z, out=op.range.element())[1]


def test_pspace_op_diagonal_call(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    r3=odl.rn(3, impl=impl, device=device)
    A = odl.IdentityOperator(r3)
    op = odl.ProductSpaceOperator([[A, 0],
                                   [0, A]])

    x = r3.element([1, 2, 3])
    y = r3.element([7, 8, 9])
    z = op.domain.element([x, y])

    assert z == op(z)
    assert z == op(z, out=op.range.element())


def test_pspace_op_swap_call(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    r3=odl.rn(3, impl=impl, device=device)
    A = odl.IdentityOperator(r3)
    op = odl.ProductSpaceOperator([[0, A],
                                   [A, 0]])

    x = r3.element([1, 2, 3])
    y = r3.element([7, 8, 9])
    z = op.domain.element([x, y])
    result = op.domain.element([y, x])

    assert result == op(z)
    assert result == op(z, out=op.range.element())


def test_comp_proj(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    r3=odl.rn(3, impl=impl, device=device)
    r3xr3 = odl.ProductSpace(r3, 2)

    x = r3xr3.element([[1, 2, 3],
                       [4, 5, 6]])
    proj_0 = odl.ComponentProjection(r3xr3, 0)
    assert x[0] == proj_0(x)
    assert x[0] == proj_0(x, out=proj_0.range.element())

    proj_1 = odl.ComponentProjection(r3xr3, 1)
    assert x[1] == proj_1(x)
    assert x[1] == proj_1(x, out=proj_1.range.element())


def test_comp_proj_slice(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    r3=odl.rn(3, impl=impl, device=device)
    r33 = odl.ProductSpace(r3, 3)

    x = r33.element([[1, 2, 3],
                     [4, 5, 6],
                     [7, 8, 9]])

    proj = odl.ComponentProjection(r33, slice(0, 2))
    assert x[0:2] == proj(x)
    assert x[0:2] == proj(x, out=proj.range.element())


def test_comp_proj_indices(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    r3=odl.rn(3, impl=impl, device=device)
    r33 = odl.ProductSpace(r3, 3)

    x = r33.element([[1, 2, 3],
                     [4, 5, 6],
                     [7, 8, 9]])

    proj = odl.ComponentProjection(r33, [0, 2])
    assert x[[0, 2]] == proj(x)
    assert x[[0, 2]] == proj(x, out=proj.range.element())


def test_comp_proj_adjoint(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    r3=odl.rn(3, impl=impl, device=device)
    r3xr3 = odl.ProductSpace(r3, 2)

    x = r3.element([1, 2, 3])

    result_0 = r3xr3.element([[1, 2, 3],
                              [0, 0, 0]])
    proj_0 = odl.ComponentProjection(r3xr3, 0)

    assert result_0 == proj_0.adjoint(x)
    assert result_0 == proj_0.adjoint(x, out=proj_0.domain.element())

    result_1 = r3xr3.element([[0, 0, 0],
                              [1, 2, 3]])
    proj_1 = odl.ComponentProjection(r3xr3, 1)

    assert result_1 == proj_1.adjoint(x)
    assert result_1 == proj_1.adjoint(x, out=proj_1.domain.element())


def test_comp_proj_adjoint_slice(odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    r3=odl.rn(3, impl=impl, device=device)
    r33 = odl.ProductSpace(r3, 3)

    x = r33[0:2].element([[1, 2, 3],
                          [4, 5, 6]])

    result = r33.element([[1, 2, 3],
                          [4, 5, 6],
                          [0, 0, 0]])
    proj = odl.ComponentProjection(r33, slice(0, 2))

    assert result == proj.adjoint(x)
    assert result == proj.adjoint(x, out=proj.domain.element())


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Convenience functions for operators."""

from __future__ import absolute_import, division, print_function

import numpy as np
from future.utils import native
from odl.core.space import ProductSpace
from odl.core.space.base_tensors import TensorSpace
from odl.core.util import nd_iterator
from odl.core.util.testutils import noise_element

__all__ = (
    'matrix_representation',
    'power_method_opnorm',
    'as_scipy_operator',
    'as_scipy_functional',
)


def matrix_representation(op):
    """Return a matrix representation of a linear operator.

    Parameters
    ----------
    op : `Operator`
        The linear operator of which one wants a matrix representation.
        If the domain or range is a `ProductSpace`, it must be a power-space.

    Returns
    -------
    matrix : `numpy.ndarray`
        The matrix representation of the operator.
        The shape will be ``op.domain.shape + op.range.shape`` and the dtype
        is the promoted (greatest) dtype of the domain and range.

    Examples
    --------
    Approximate a matrix on its own:

    >>> mat = np.array([[1, 2, 3],
    ...                 [4, 5, 6],
    ...                 [7, 8, 9]])
    >>> op = odl.MatrixOperator(mat)
    >>> matrix_representation(op)
    array([[1, 2, 3],
           [4, 5, 6],
           [7, 8, 9]])

    It also works with `ProductSpace`'s and higher dimensional `TensorSpace`'s.
    In this case, the returned "matrix" will also be higher dimensional:

    >>> space = odl.uniform_discr([0, 0], [2, 2], (2, 2))
    >>> grad = odl.Gradient(space)
    >>> tensor = odl.matrix_representation(grad)
    >>> tensor.shape == (2, 2, 2, 2, 2)
    True

    Since the "matrix" is now higher dimensional, we need to use e.g.
    `numpy.tensordot` if we want to compute with the matrix representation:

    >>> x = space.element(lambda x: x[0] ** 2 + 2 * x[1] ** 2)
    >>> grad(x)
    ProductSpace(uniform_discr([ 0.,  0.], [ 2.,  2.], (2, 2)), 2).element([
    <BLANKLINE>
            [[ 2.  ,  2.  ],
             [-2.75, -6.75]],
    <BLANKLINE>
            [[ 4.  , -4.75],
             [ 4.  , -6.75]]
    ])
    >>> np.tensordot(tensor, x.data, axes=grad.domain.ndim)
    array([[[ 2.  ,  2.  ],
            [-2.75, -6.75]],
    <BLANKLINE>
           [[ 4.  , -4.75],
            [ 4.  , -6.75]]])

    Notes
    ----------
    The algorithm works by letting the operator act on all unit vectors, and
    stacking the output as a matrix.
    """

    if not op.is_linear:
        raise ValueError('the operator is not linear')

    if not (isinstance(op.domain, TensorSpace) or
            (isinstance(op.domain, ProductSpace) and
             op.domain.is_power_space and
             all(isinstance(spc, TensorSpace) for spc in op.domain))):
        raise TypeError('operator domain {!r} is neither `TensorSpace` '
                        'nor `ProductSpace` with only equal `TensorSpace` '
                        'components'.format(op.domain))

    if not (isinstance(op.range, TensorSpace) or
            (isinstance(op.range, ProductSpace) and
             op.range.is_power_space and
             all(isinstance(spc, TensorSpace) for spc in op.range))):
        raise TypeError('operator range {!r} is neither `TensorSpace` '
                        'nor `ProductSpace` with only equal `TensorSpace` '
                        'components'.format(op.range))

    # Generate the matrix
    if isinstance(op.domain, TensorSpace):
        namespace = op.domain.array_namespace
        device    = op.domain.device
    else:
        namespace = op[0][0].domain.array_namespace
        device    = op[0][0].domain.device
    dtype  = namespace.result_type(op.domain.dtype, op.range.dtype)
    matrix = namespace.zeros(
        op.range.shape + op.domain.shape, dtype=dtype, device=device)
    tmp_ran = op.range.element()  # Store for reuse in loop
    tmp_dom = op.domain.zero()  # Store for reuse in loop

    for j in nd_iterator(op.domain.shape):
        tmp_dom[j] = 1.0

        op(tmp_dom, out=tmp_ran)
        matrix[(Ellipsis,) + j] = tmp_ran.asarray()

        tmp_dom[j] = 0.0

    return matrix


def power_method_opnorm(op, xstart=None, maxiter=100, rtol=1e-05, atol=1e-08,
                        callback=None):
    r"""Estimate the operator norm with the power method.

    Parameters
    ----------
    op : `Operator`
        Operator whose norm is to be estimated. If its `Operator.range`
        range does not coincide with its `Operator.domain`, an
        `Operator.adjoint` must be defined (which implies that the
        operator must be linear).
    xstart : ``op.domain`` `element-like`, optional
        Starting point of the iteration. By default an `Operator.domain`
        element containing noise is used.
    maxiter : positive int, optional
        Number of iterations to perform. If the domain and range of ``op``
        do not match, it needs to be an even number. If ``None`` is given,
        iterate until convergence.
    rtol : float, optional
        Relative tolerance parameter (see Notes).
    atol : float, optional
        Absolute tolerance parameter (see Notes).
    callback : callable, optional
        Function called with the current iterate in each iteration.

    Returns
    -------
    est_opnorm : float
        The estimated operator norm of ``op``.

    Examples
    --------
    Verify that the identity operator has norm close to 1:

    >>> space = odl.uniform_discr(0, 1, 5)
    >>> id = odl.IdentityOperator(space)
    >>> estimation = power_method_opnorm(id)
    >>> np.round(estimation, decimals=3)
    1.0

    Notes
    -----
    The operator norm :math:`||A||` is defined by as the smallest number
    such that

    .. math::
        ||A(x)|| \leq ||A|| ||x||

    for all :math:`x` in the domain of :math:`A`.

    The operator is evaluated until ``maxiter`` operator calls or until the
    relative error is small enough. The error measure is given by

        ``abs(a - b) <= (atol + rtol * abs(b))``,

    where ``a`` and ``b`` are consecutive iterates.
    """
    if maxiter is None:
        maxiter = np.iinfo(int).max

    maxiter, maxiter_in = int(maxiter), maxiter
    if maxiter <= 0:
        raise ValueError('`maxiter` must be positive, got {}'
                         ''.format(maxiter_in))

    if op.adjoint is op:
        use_normal = False
        ncalls = maxiter
    else:
        # Do the power iteration for A*A; the norm of A*A(x_N) is then
        # an estimate of the square of the operator norm
        # We do only half the number of iterations compared to the usual
        # case to have the same number of operator evaluations.
        use_normal = True
        ncalls = maxiter // 2
        if ncalls * 2 != maxiter:
            raise ValueError('``maxiter`` must be an even number for '
                             'non-self-adjoint operator, got {}'
                             ''.format(maxiter_in))

    # Make sure starting point is ok or select initial guess
    if xstart is None:
        x = noise_element(op.domain)
    else:
        # copy to ensure xstart is not modified
        x = op.domain.element(xstart).copy()

    # Take first iteration step to normalize input
    x_norm = x.norm()
    if x_norm == 0:
        raise ValueError('``xstart`` must be nonzero')
    x /= x_norm

    # utility to calculate opnorm from xnorm
    def calc_opnorm(x_norm):
        if use_normal:
            return np.sqrt(x_norm)
        else:
            return x_norm

    # initial guess of opnorm
    opnorm = calc_opnorm(x_norm)

    # temporary to improve performance
    tmp = op.range.element()

    # Use the power method to estimate opnorm
    for i in range(ncalls):
        if use_normal:
            op(x, out=tmp)
            op.adjoint(tmp, out=x)
        else:
            op(x, out=tmp)
            x, tmp = tmp, x

        # Calculate x norm and verify it is valid
        x_norm = x.norm()
        if x_norm == 0:
            raise ValueError('reached ``x=0`` after {} iterations'.format(i))
        if not np.isfinite(x_norm):
            raise ValueError('reached nonfinite ``x={}`` after {} iterations'
                             ''.format(x, i))

        # Calculate opnorm
        opnorm, opnorm_old = calc_opnorm(x_norm), opnorm

        # If the breaking condition holds, stop. Else rescale and go on.
        if np.isclose(opnorm, opnorm_old, rtol, atol):
            break
        else:
            x /= x_norm

        if callback is not None:
            callback(x)

    return opnorm


def as_scipy_operator(op):
    """Wrap ``op`` as a ``scipy.sparse.linalg.LinearOperator``.

    This is intended to be used with the scipy sparse linear solvers.

    Parameters
    ----------
    op : `Operator`
        A linear operator that should be wrapped

    Returns
    -------
    ``scipy.sparse.linalg.LinearOperator`` : linear_op
        The wrapped operator, has attributes ``matvec`` which calls ``op``,
        and ``rmatvec`` which calls ``op.adjoint``.

    Examples
    --------
    Wrap operator and solve simple problem (here toy problem ``Ix = b``)

    >>> op = odl.IdentityOperator(odl.rn(3))
    >>> scipy_op = as_scipy_operator(op)
    >>> import scipy.sparse.linalg as scipy_solvers
    >>> result, status = scipy_solvers.cg(scipy_op, [0, 1, 0])
    >>> result
    array([ 0.,  1.,  0.])

    Notes
    -----
    If the data representation of ``op``'s domain and range is of type
    `NumpyTensorSpace` this incurs no significant overhead. If the space
    type is ``CudaFn`` or some other nonlocal type, the overhead is
    significant.
    """
    # Lazy import to improve `import odl` time
    import scipy.sparse

    if not op.is_linear:
        raise ValueError('`op` needs to be linear')

    dtype = op.domain.dtype
    if op.range.dtype != dtype:
        raise ValueError('dtypes of ``op.domain`` and ``op.range`` needs to '
                         'match')

    shape = (native(op.range.size), native(op.domain.size))

    def matvec(v):
        return (op(v.reshape(op.domain.shape))).asarray().ravel()

    def rmatvec(v):
        return (op.adjoint(v.reshape(op.range.shape))).asarray().ravel()

    return scipy.sparse.linalg.LinearOperator(shape=shape,
                                              matvec=matvec,
                                              rmatvec=rmatvec,
                                              dtype=dtype)


def as_scipy_functional(func, return_gradient=False):
    """Wrap ``op`` as a function operating on linear arrays.

    This is intended to be used with the `scipy solvers
    <https://docs.scipy.org/doc/scipy/reference/optimize.html>`_.

    Parameters
    ----------
    func : `Functional`.
        A functional that should be wrapped
    return_gradient : bool, optional
        ``True`` if the gradient of the functional should also be returned,
        ``False`` otherwise.

    Returns
    -------
    function : ``callable``
        The wrapped functional.
    gradient : ``callable``, optional
        The wrapped gradient. Only returned if ``return_gradient`` is true.

    Examples
    --------
    Wrap functional and solve simple problem
    (here toy problem ``min_x ||x||^2``):

    >>> func = odl.functionals.L2NormSquared(odl.rn(3))
    >>> scipy_func = odl.as_scipy_functional(func)
    >>> from scipy.optimize import minimize
    >>> result = minimize(scipy_func, x0=[0, 1, 0])
    >>> np.allclose(result.x, [0, 0, 0])
    True

    The gradient (jacobian) can also be provided:

    >>> func = odl.functionals.L2NormSquared(odl.rn(3))
    >>> scipy_func, scipy_grad = odl.as_scipy_functional(func, True)
    >>> from scipy.optimize import minimize
    >>> result = minimize(scipy_func, x0=[0, 1, 0], jac=scipy_grad)
    >>> np.allclose(result.x, [0, 0, 0])
    True

    Notes
    -----
    If the data representation of ``op``'s domain is of type
    `NumpyTensorSpace`, this incurs no significant overhead. If the space type
    is ``CudaFn`` or some other nonlocal type, the overhead is significant.
    """
    def func_call(arr):
        return func(func.domain.element(np.asarray(arr).reshape(func.domain.shape)))

    if return_gradient:
        def func_gradient_call(arr):
            return np.asarray(
                func.gradient(np.asarray(arr).reshape(func.domain.shape)).data)

        return func_call, func_gradient_call
    else:
        return func_call


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests

    run_doctests()



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Solvers for non-smooth optimization problems."""

from __future__ import absolute_import

from .admm import *
from .alternating_dual_updates import *
from .difference_convex import *
from .douglas_rachford import *
from .forward_backward import *
from .primal_dual_hybrid_gradient import *
from .proximal_gradient_solvers import *
from .proximal_operators import *

__all__ = ()
__all__ += admm.__all__
__all__ += alternating_dual_updates.__all__
__all__ += difference_convex.__all__
__all__ += douglas_rachford.__all__
__all__ += forward_backward.__all__
__all__ += primal_dual_hybrid_gradient.__all__
__all__ += proximal_gradient_solvers.__all__
__all__ += proximal_operators.__all__



# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import division
import numpy as np
import pytest

import odl
from odl.trafos.backends import pyfftw_call, PYFFTW_AVAILABLE
from odl.core.util import (
    is_real_dtype, complex_dtype)
from odl.core.util.testutils import (
    all_almost_equal, simple_fixture)
from odl.core.util.dtype_utils import FLOAT_DTYPES, COMPLEX_DTYPES

pytestmark = pytest.mark.skipif(not PYFFTW_AVAILABLE,
                                reason='`pyfftw` backend not available')


# --- pytest fixtures --- #


planning = simple_fixture('planning', ['estimate', 'measure', 'patient',
                                       'exhaustive'])
direction = simple_fixture('direction', ['forward', 'backward'])


# --- helper functions --- #


def _random_array(shape, dtype):
    if is_real_dtype(dtype):
        return np.random.rand(*shape).astype(dtype)
    else:
        return (np.random.rand(*shape).astype(dtype) +
                1j * np.random.rand(*shape).astype(dtype))


def _params_from_dtype(dtype):
    if is_real_dtype(dtype):
        halfcomplex = True
    else:
        halfcomplex = False
    return halfcomplex, complex_dtype(dtype)


def _halfcomplex_shape(shape, axes=None):
    if axes is None:
        axes = tuple(range(len(shape)))

    try:
        axes = (int(axes),)
    except TypeError:
        pass

    shape = list(shape)
    shape[axes[-1]] = shape[axes[-1]] // 2 + 1
    return shape


# ---- pyfftw_call ---- #


def test_pyfftw_call_forward(odl_floating_dtype):
    # Test against Numpy's FFT
    dtype = odl_floating_dtype
    if dtype == np.dtype('float16'):  # not supported, skipping
        return

    halfcomplex, out_dtype = _params_from_dtype(dtype)

    for shape in [(10,), (3, 4, 5)]:
        arr = _random_array(shape, dtype)

        if halfcomplex:
            true_dft = np.fft.rfftn(arr)
            dft_arr = np.empty(_halfcomplex_shape(shape), dtype=out_dtype)
        else:
            true_dft = np.fft.fftn(arr)
            dft_arr = np.empty(shape, dtype=out_dtype)

        pyfftw_call(arr, dft_arr, direction='forward',
                    halfcomplex=halfcomplex, preserve_input=False)

        assert all_almost_equal(dft_arr, true_dft)


def test_pyfftw_call_threads():
    shape = (3, 4, 5)
    arr = _random_array(shape, dtype='complex64')
    true_dft = np.fft.fftn(arr)
    dft_arr = np.empty(shape, dtype='complex64')
    pyfftw_call(arr, dft_arr, direction='forward', preserve_input=False,
                threads=4)
    assert all_almost_equal(dft_arr, true_dft)

    shape = (1000,)  # Trigger cpu_count() as number of threads
    arr = _random_array(shape, dtype='complex64')
    true_dft = np.fft.fftn(arr)
    dft_arr = np.empty(shape, dtype='complex64')
    pyfftw_call(arr, dft_arr, direction='forward', preserve_input=False)

    assert all_almost_equal(dft_arr, true_dft)


def test_pyfftw_call_backward(odl_floating_dtype):
    # Test against Numpy's IFFT, no normalization
    dtype = odl_floating_dtype
    if dtype == np.dtype('float16'):  # not supported, skipping
        return

    halfcomplex, in_dtype = _params_from_dtype(dtype)

    for shape in [(10,), (3, 4, 5)]:
        # Scaling happens wrt output (large) shape
        idft_scaling = np.prod(shape)

        if halfcomplex:
            arr = _random_array(_halfcomplex_shape(shape), in_dtype)
            true_idft = np.fft.irfftn(arr, shape) * idft_scaling
        else:
            arr = _random_array(shape, in_dtype)
            true_idft = np.fft.ifftn(arr) * idft_scaling

        idft_arr = np.empty(shape, dtype=dtype)
        pyfftw_call(arr, idft_arr, direction='backward',
                    halfcomplex=halfcomplex)

        assert all_almost_equal(idft_arr, true_idft)


def test_pyfftw_call_bad_input(direction):

    # Complex

    # Bad dtype
    dtype_in = np.dtype('complex128')
    arr_in = np.empty(3, dtype=dtype_in)
    backend = odl.lookup_array_backend('numpy')
    float_dt = {backend.available_dtypes[dtype] for dtype in FLOAT_DTYPES}
    complex_dt = {backend.available_dtypes[dtype] for dtype in COMPLEX_DTYPES}
    bad_dtypes_out = float_dt.union( complex_dt)
    if dtype_in in bad_dtypes_out:
        # This one is correct, so we remove it
        bad_dtypes_out.remove(dtype_in)

    for bad_dtype in bad_dtypes_out:
        arr_out = np.empty(3, dtype=bad_dtype)
        with pytest.raises(ValueError):
            pyfftw_call(arr_in, arr_out, halfcomplex=False,
                        direction=direction)

    # Bad shape
    shape = (3, 4)
    arr_in = np.empty(shape, dtype='complex128')
    bad_shapes_out = [(3, 3), (3,), (4,), (3, 4, 5), ()]
    for bad_shape in bad_shapes_out:
        arr_out = np.empty(bad_shape, dtype='complex128')
        with pytest.raises(ValueError):
            pyfftw_call(arr_in, arr_out, halfcomplex=False,
                        direction=direction)

    # Duplicate axes
    arr_in = np.empty((3, 4, 5), dtype='complex128')
    arr_out = np.empty_like(arr_in)
    bad_axes_list = [(0, 0, 1), (1, 1, 1), (-1, -1)]
    for bad_axes in bad_axes_list:
        with pytest.raises(ValueError):
            pyfftw_call(arr_in, arr_out, axes=bad_axes,
                        direction=direction)

    # Axis entry out of range
    arr_in = np.empty((3, 4, 5), dtype='complex128')
    arr_out = np.empty_like(arr_in)
    bad_axes_list = [(0, 3), (-4,)]
    for bad_axes in bad_axes_list:
        with pytest.raises(ValueError):
            pyfftw_call(arr_in, arr_out, axes=bad_axes,
                        direction=direction)

    # Halfcomplex not possible for complex data
    arr_in = np.empty((3, 4, 5), dtype='complex128')
    arr_out = np.empty_like(arr_in)
    with pytest.raises(ValueError):
        pyfftw_call(arr_in, arr_out, halfcomplex=True,
                    direction=direction)

    # Data type mismatch
    arr_in = np.empty((3, 4, 5), dtype='complex128')
    arr_out = np.empty_like(arr_in, dtype='complex64')
    with pytest.raises(ValueError):
        pyfftw_call(arr_in, arr_out, direction=direction)

    # Halfcomplex

    # Bad dtype
    dtype_in = 'float64'
    arr_in = np.empty(10, dtype=dtype_in)
    backend = odl.lookup_array_backend('numpy')
    float_dt = {backend.available_dtypes[dtype] for dtype in FLOAT_DTYPES}
    complex_dt = {backend.available_dtypes[dtype] for dtype in COMPLEX_DTYPES}
    bad_dtypes_out = float_dt.union( complex_dt)
    try:
        # This one is correct, so we remove it
        bad_dtypes_out.remove(np.dtype('complex128'))
    except ValueError:
        pass
    for bad_dtype in bad_dtypes_out:
        arr_out = np.empty(6, dtype=bad_dtype)
        with pytest.raises(ValueError):
            if direction == 'forward':
                pyfftw_call(arr_in, arr_out, halfcomplex=True,
                            direction='forward')
            else:
                pyfftw_call(arr_out, arr_in, halfcomplex=True,
                            direction='backward')

    # Bad shape
    shape = (3, 4, 5)
    axes_list = [None, (0, 1), (1,), (1, 2), (2, 1), (-1, -2, -3)]
    arr_in = np.empty(shape, dtype='float64')
    # Correct shapes:
    # [(3, 4, 3), (3, 3, 5), (3, 3, 5), (3, 4, 3), (3, 3, 5), (2, 4, 5)]
    bad_shapes_out = [(3, 4, 2), (3, 4, 3), (2, 3, 5), (3, 2, 3),
                      (3, 4, 3), (3, 4, 3)]
    always_bad_shapes = [(3, 4), (3, 4, 5)]
    for bad_shape, axes in zip(bad_shapes_out, axes_list):

        for always_bad_shape in always_bad_shapes:
            arr_out = np.empty(always_bad_shape, dtype='complex128')
            with pytest.raises(ValueError):
                if direction == 'forward':
                    pyfftw_call(arr_in, arr_out, axes=axes, halfcomplex=True,
                                direction='forward')
                else:
                    pyfftw_call(arr_out, arr_in, axes=axes, halfcomplex=True,
                                direction='backward')

        arr_out = np.empty(bad_shape, dtype='complex128')
        with pytest.raises(ValueError):
            if direction == 'forward':
                pyfftw_call(arr_in, arr_out, axes=axes, halfcomplex=True,
                            direction='forward')
            else:
                pyfftw_call(arr_out, arr_in, axes=axes, halfcomplex=True,
                            direction='backward')


def test_pyfftw_call_forward_real_not_halfcomplex():
    # Test against Numpy's FFT
    for shape in [(10,), (3, 4, 5)]:
        arr = _random_array(shape, dtype='float64')

        true_dft = np.fft.fftn(arr)
        dft_arr = np.empty(shape, dtype='complex128')
        pyfftw_call(arr, dft_arr, direction='forward', halfcomplex=False)

        assert all_almost_equal(dft_arr, true_dft)


def test_pyfftw_call_backward_real_not_halfcomplex():
    # Test against Numpy's IFFT, no normalization
    for shape in [(10,), (3, 4, 5)]:
        # Scaling happens wrt output (large) shape
        idft_scaling = np.prod(shape)

        arr = _random_array(shape, dtype='float64')
        true_idft = np.fft.ifftn(arr) * idft_scaling
        idft_arr = np.empty(shape, dtype='complex128')
        pyfftw_call(arr, idft_arr, direction='backward', halfcomplex=False)

        assert all_almost_equal(idft_arr, true_idft)


def test_pyfftw_call_plan_preserve_input(planning):

    for shape in [(10,), (3, 4)]:
        arr = _random_array(shape, dtype='complex128')
        arr_cpy = arr.copy()

        idft_scaling = np.prod(shape)
        true_idft = np.fft.ifftn(arr) * idft_scaling
        idft_arr = np.empty(shape, dtype='complex128')
        pyfftw_call(arr, idft_arr, direction='backward', halfcomplex=False,
                    planning=planning)

        assert all_almost_equal(arr, arr_cpy)  # Input perserved
        assert all_almost_equal(idft_arr, true_idft)


def test_pyfftw_call_forward_with_axes(odl_floating_dtype):
    dtype = odl_floating_dtype
    if dtype == np.dtype('float16'):  # not supported, skipping
        return

    halfcomplex, out_dtype = _params_from_dtype(dtype)
    shape = (3, 4, 5)

    test_axes = [(0, 1), [1], (-1,), (1, 0), (-1, -2, -3)]
    for axes in test_axes:
        arr = _random_array(shape, dtype)
        if halfcomplex:
            true_dft = np.fft.rfftn(arr, axes=axes)
            dft_arr = np.empty(_halfcomplex_shape(shape, axes),
                               dtype=out_dtype)
        else:
            true_dft = np.fft.fftn(arr, axes=axes)
            dft_arr = np.empty(shape, dtype=out_dtype)

        pyfftw_call(arr, dft_arr, direction='forward', axes=axes,
                    halfcomplex=halfcomplex)

        assert all_almost_equal(dft_arr, true_dft)


def test_pyfftw_call_backward_with_axes(odl_floating_dtype):
    dtype = odl_floating_dtype
    if dtype == np.dtype('float16'):  # not supported, skipping
        return

    halfcomplex, in_dtype = _params_from_dtype(dtype)
    shape = (3, 4, 5)

    test_axes = [(0, 1), [1], (-1,), (1, 0), (-1, -2, -3)]
    for axes in test_axes:
        # Only the shape indexed by axes count for the scaling
        active_shape = np.take(shape, axes)
        idft_scaling = np.prod(active_shape)

        if halfcomplex:
            arr = _random_array(_halfcomplex_shape(shape, axes), in_dtype)
            true_idft = (np.fft.irfftn(arr, s=active_shape, axes=axes) *
                         idft_scaling)
        else:
            arr = _random_array(shape, in_dtype)
            true_idft = (np.fft.ifftn(arr, s=active_shape, axes=axes) *
                         idft_scaling)

        idft_arr = np.empty(shape, dtype=dtype)
        pyfftw_call(arr, idft_arr, direction='backward', axes=axes,
                    halfcomplex=halfcomplex)

        assert all_almost_equal(idft_arr, true_idft)


def test_pyfftw_call_forward_with_plan():

    for shape in [(10,), (3, 4, 5)]:
        arr = _random_array(shape, dtype='complex128')
        arr_cpy = arr.copy()
        true_dft = np.fft.fftn(arr)

        # First run, create plan
        dft_arr = np.empty(shape, dtype='complex128')
        plan = pyfftw_call(arr, dft_arr, direction='forward',
                           halfcomplex=False, planning_effort='measure')

        # Second run, reuse with fresh output array
        dft_arr = np.empty(shape, dtype='complex128')
        pyfftw_call(arr, dft_arr, direction='forward', fftw_plan=plan,
                    halfcomplex=False)

        assert all_almost_equal(arr, arr_cpy)  # Input perserved
        assert all_almost_equal(dft_arr, true_dft)


def test_pyfftw_call_backward_with_plan():

    for shape in [(10,), (3, 4, 5)]:
        arr = _random_array(shape, dtype='complex128')
        arr_cpy = arr.copy()
        idft_scaling = np.prod(shape)
        true_idft = np.fft.ifftn(arr) * idft_scaling

        # First run, create plan
        idft_arr = np.empty(shape, dtype='complex128')
        plan = pyfftw_call(arr, idft_arr, direction='backward',
                           halfcomplex=False, planning_effort='measure')

        # Second run, reuse with fresh output array
        idft_arr = np.empty(shape, dtype='complex128')
        pyfftw_call(arr, idft_arr, direction='backward', fftw_plan=plan,
                    halfcomplex=False)

        assert all_almost_equal(arr, arr_cpy)  # Input perserved
        assert all_almost_equal(idft_arr, true_idft)


if __name__ == '__main__':
    odl.core.util.test_file(__file__)
